#ifndef _MMEEP_WRAPPER_H_
#define _MMEEP_WRAPPER_H_

#include <assert.h>
#include <stdio.h>
#include <vector>
#include <utility>
#include <functional>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <cstring>
#include <complex>

#include <meep.hpp>


using namespace meep;


////////////////////////////////////////////////////////////////////////////////
//
// MMEEP_WRAPPER
//
////////////////////////////////////////////////////////////////////////////////

class MMEEP_WRAPPER {
 public:
    int _id;
 
    // () parameters that are different for each different instance of meep simulation;
    double _patch_length; // patchL; param_values(0,0);
    double _patch_width; // patchW; param_values(1,0);
    double _recess; // param_values(2,0);
    double _depth; // param_values(3,0);


    // () variables that Jim had as globals all over the place;
    // main parameters: 6 of them;
    double blockCm;
    double stripWidthBlk;
    // four of them will be having values as dictated by user via input file;
    // but, note that their values will be computed indirectly; 
    double patchLBlk;
    double patchWBlk;
    double rBlk;
    double dBlk;

    double stripThickBlk;
    double structWidthBlk;
    double WLblk;
    double freqBlk,dfBlk;
    double XmaxBlk;
    double YmaxBlk;
    double portBlk;
    double sourceBlk;
    double AddTlBlk;
    double topSpaceBlk;
    double PmlThickBlk;

    double resolution;
    int nfreq;
    double epsR;
    double sx,sy,sz;
    double zUpAir,zUpEps;

    double *flux1t, *flux1r;
    std::complex<double> *Ezt1, *Hyt1, *Ezr1, *Hyr1;
    std::complex<double> *Ezt2, *Hyt2, *Ezr2, *Hyr2;
    int NumPts;

    // overall "cost" of this solution; used for the purpose of simulated annealing
    double _cost;
    double _cost_c1, _cost_c2;

    // sketch variables to store result of meep simulation; these used to be
    // local variables inside findSpar();
    double *_frequency;
    double *_S11;
    int _iFreqMin;
    double _center_freq;
    double _BW;
    

 public:
    MMEEP_WRAPPER(int id, double patch_length, double patch_width, double recess, double depth);
    ~MMEEP_WRAPPER() {
        // clean up; 
        delete flux1r; // allocated inside read_user_refData();
        delete _frequency; // allocated inside Ctor;
        delete _S11; // allocated inside Ctor;
    }
    
    void set_patch_length(double val) { _patch_length = val; }
    void set_patch_width(double val) { _patch_width = val; }
    void set_recess(double val) { _recess = val; }
    void set_depth(double val) { _depth = val; }
    int get_id() { return _id; }
    double patch_length() { return _patch_length; }
    double patch_width() { return _patch_width; }
    double recess() { return _recess; }
    double depth() { return _depth; }
    double cost() { return _cost; }
    double cost_c1() { return _cost_c1; }
    double cost_c2() { return _cost_c2; }
    double center_freq() { return _center_freq; }
    double BW() { return _BW; }
    int get_nfreq() { return nfreq; }
    double frequency(int i) { return _frequency[i]; }
    double S11(int i) { return _S11[i]; }

    // () initializations;
    // initialize geometries based on the input parameters; 
    // called only once at the beginning;
    void initialize_geometrical_paramaters();
    // problem statement; more params; also called only once;
    void problem_statement_paramaters();
    void read_user_refData();
    void print_parameters_materials(); // Jim's initial void printParameters(void);

    // () these functions described in materials.cpp
    // \cmt{Functions to define the material properties}
    // \cmt{Reference Line: used in both Resistor and QWT problems}
    // Note: I make these functions static to be able to use them in the Ctor
    // of structure inside runMeep() function;
    //double patch(const vec &p);
    // \cmt{Loss function, for both patch and ref. line}
    //double Loss(const vec &p);
    // \cmt{Test line, material properties (reistor problem only)}
    //double RefLine(const vec &p);

    // () function described in runMeep.cpp
    void runMEEP(void);
    // () function described in findSpar.cpp
    void findSpar(void);


    // runMEEP() and findSpar() combined; calculates cost of this solution
    // for SA purposes;
    void do_meep_simulation_to_evaluate_solution();
    double calculate_cost_function_for_SA();


    // () Note: this class replaces jim's initial "double patch(const vec &p)"
    // that he was passing to the ctor of structure inside runMeep();
    // it is derived from meep's "class material_function";
    class patch_material_function : public material_function {
    public:
        MMEEP_WRAPPER *_m_w;
        double _core_f(const vec &p) {
            double x,y,z,xps,xpr,xpe,yc;
            x=p.x();
            y=p.y();
            z=p.z();
            yc = _m_w->sy/2.;
            xps = _m_w->PmlThickBlk + _m_w->AddTlBlk + _m_w->sourceBlk + _m_w->portBlk;
            xpr = xps + _m_w->rBlk;
            xpe = xps + _m_w->patchLBlk;

            if (z < _m_w->zUpEps)
                return _m_w->epsR;
            if (z > _m_w->zUpAir)
                return 1.;
            if ( (z >= _m_w->zUpEps) && (z <= _m_w->zUpAir) ) {
                if (x < xps) {
                    if (fabs(y - yc) <= _m_w->stripWidthBlk/2.)
                        return -1.*infinity;
                    if (fabs(y - yc) > _m_w->stripWidthBlk/2.)
                        return 1.;
                }
                if ( (x >= xps) && (x < xpr) ) {
                    if ( (fabs(y-yc) <= _m_w->stripWidthBlk/2.) ||
                        ( ( fabs(y-yc) > _m_w->stripWidthBlk/2. + _m_w->dBlk) && 
                        (fabs(y-yc) <= _m_w->patchWBlk/2.) ))
                        return -1.*infinity;
                    if ( (fabs(y-yc) > _m_w->patchWBlk/2.) ||
                        ( ( fabs(y-yc) > _m_w->stripWidthBlk/2.) &&
                        (fabs(y-yc) <= _m_w->stripWidthBlk/2. + _m_w->dBlk)))
                        return 1.;
                }
                if ( (x >= xpr) && (x < xpe)) {
                    if (fabs(y-yc) <= _m_w->patchWBlk/2.)
                        return -1.*infinity;
                    if (fabs(y-yc) > _m_w->patchWBlk/2.)
                        return 1.;
                }
                if (x>=xpe)
                    return 1.;
            }
            // \cmt{exit if no value is assigned}
            //master_printf("in referenceLine\n"); 
            //exit(1);
        }
    public:
        patch_material_function(MMEEP_WRAPPER *mmeep_wrapper) { _m_w = mmeep_wrapper;}
        ~patch_material_function() {}
  
        double chi1p1(field_type ft, const vec &r) { (void)ft; return _core_f(r); }
        double eps(const vec &r) { return _core_f(r); }
        double mu(const vec &r) { return _core_f(r); }
        double conductivity(component c, const vec &r) { 
            (void)c; return _core_f(r); }
        void sigma_row(component c, double sigrow[3], const vec &r) {
            sigrow[0] = sigrow[1] = sigrow[2] = 0.0;
            sigrow[component_index(c)] = _core_f(r);
        }
        double chi3(component c, const vec &r) { (void)c; return _core_f(r); }
        double chi2(component c, const vec &r) { (void)c; return _core_f(r); }
    };


    // () Note: this class replaces jim's initial "double Loss(const vec &p)"
    // that he was passing to s->set_conductivity() inside runMeep();
    // it is derived from meep's "class material_function";
    class loss_material_function : public material_function {
    public:
        MMEEP_WRAPPER *_m_w;
        double _core_f(const vec &p) {
            double z;
            double sigma=1.086e-3;
            z = p.z();
            if (z < _m_w->zUpEps)
                return sigma;
            if (z > _m_w->zUpAir)
                return 0;
            if ( (z >= _m_w->zUpEps) && (z <= _m_w->zUpAir) )
                return 0.;
            // \cmt{exit if no value assigned}
            //master_printf("in Loss\n");
            //exit(1);
        }
    public:
        loss_material_function(MMEEP_WRAPPER *mmeep_wrapper) { _m_w = mmeep_wrapper;}
        ~loss_material_function() {}
  
        double chi1p1(field_type ft, const vec &r) { (void)ft; return _core_f(r); }
        double eps(const vec &r) { return _core_f(r); }
        double mu(const vec &r) { return _core_f(r); }
        double conductivity(component c, const vec &r) { 
            (void)c; return _core_f(r); }
        void sigma_row(component c, double sigrow[3], const vec &r) {
            sigrow[0] = sigrow[1] = sigrow[2] = 0.0;
            sigrow[component_index(c)] = _core_f(r);
        }
        double chi3(component c, const vec &r) { (void)c; return _core_f(r); }
        double chi2(component c, const vec &r) { (void)c; return _core_f(r); }
    };

};

#endif


