#include "mtpc_app.h"

using namespace std;


DIRECTOR g_director;
MUTEX global_mutex("global_mutex_of_MTPC_class");
static SIGNAL_TIMER s_global_signal_timer;


////////////////////////////////////////////////////////////////////////////////
//
// MTPC
//
////////////////////////////////////////////////////////////////////////////////


MTPC::MTPC(APPLICATION_SIMULATED_ANNEALING_DSE *application) :
    _application(application)
{
    // create controller object;
    _controller = new CONTROLLER(this);
    // record pointer to this host inside the application that will be run;
    _application->set_mtpc(this);
    // record cpus count from global director;
    _cpus_count = max(1, g_director.get_cpus_count());
    assert(_cpus_count < INITIAL_VECTOR_SIZE);
    _tasks.resize(INITIAL_VECTOR_SIZE, 0);
    _tasks_count = 0;
}


MTPC::~MTPC()
{
    for (int i = 0; i < _tasks_count; i++) {
        if (_tasks[i]) delete _tasks[i];
    }
}


MTPC::TASK *MTPC::add_task(MMEEP_WRAPPER *mmeep_wrapper, int priority, string &name)
{
    // create new task with its state as INITIAL; add to list of tasks;
    TASK *task = new TASK(mmeep_wrapper, priority, name);
    assert(task);
    task->set_id( _tasks_count);
    _tasks[ _tasks_count ++] = task;
    return task;
}


int MTPC::run()
{
    g_director.print_director_initial();

    // () generate an initial number of solution points to start with;
    int tasks_count = _application->create_initial_tasks();
    printf("\nApplication created initial tasks: %d ", tasks_count);


    // () controller keeps track of status; initialization done here as
    // it's minimal (no runtime); should revisit its location if more
    // generators are used! this does part of sync;
    _controller->initialize( _tasks_count, _cpus_count);
    printf("\n --- Application state-machine controller : initialized");


    if ( _cpus_count == 1) { // run_sequentialy, yet in the order of priorities;
        DATA *data;
        while (data = run_generator_thread_one()) {
            run_worker_thread_one(data);
            run_transfer_thread_one(data);
        }
    } else {
        // create generators and workers; their creation is directed
        // by the global "director";


        // (a) create one generator custom thread;
        ONE_THREAD *generator = ONE_THREAD::create_thread(this);
        // start it by actually creating a pthread to which function 
        // generator_thread_start() is passed; this function in turn, runs
        // mtpc's run_generator_thread(); this last one, creates data and places 
        // in _main_queue; data is created by picking up the best next task,
        // from the vector of tasks; picking up next best task is done by calling
        // in turn mtpc's run_generator_thread_one();
        if (generator->start_as_generator() != 0) { // <--- key1
            delete generator;
            return 0; // problem!
        }
        printf("\n --- Created and launched one generator thread");


        // (b) create worker "custom threads" (e.g., ONE_THREAD); we create as 
        // many worker threads as cpu cores on the machine that we are executing;
        vector<ONE_THREAD *> workers;
        workers.resize(INITIAL_VECTOR_SIZE, 0);
        int i = 0;
        int counter = 0;
        for (i = 0; i < _cpus_count; i++) {
            workers[ i] = ONE_THREAD::create_thread(this);
            // start it by actually creating a pthread to which function 
            // worker_thread_start() is passed; this, in turn, runs 
            // mtpc's run_worker_thread(), which, again in turn, 
            // begins to get data objects from main_queue, one by one as
            // long as they exist (note, all worker threads consume and do this)
            // and executes function run_worker_thread_one( data) on them;
            if (workers[i]->start_as_worker() == 0) { // <--- key2
                // () account for this successfully created worker thread;
                counter++;
                // () I want to have each of the workers scheduled to a separate 
                // CPU core; use worker id to select the core index;
                int rc = workers[i]->set_affinity_custom( i);
                if (rc != 0) {
                    std::cerr << "Error calling pthread_setaffinity_np: " << rc << "\n";
                }
            }
        }
        printf("\n --- Created and launched %d worker threads", counter);
        if (counter == 0) {
            for (i = 0; i < _cpus_count; i++) delete workers[i];
            return 0; // problem!
        }


        // (c) wait for all generator(s);
        wait_generators(1); 
        _main_queue.set_stop();
        wait_workers(_cpus_count);
        delete generator;
        for (i = 0; i < _cpus_count; i++) {
            delete workers[i];
        }
    }

    return 1;
}


////////////////////////////////////////////////////////////////////////////////
//
// functions related to generators
//
////////////////////////////////////////////////////////////////////////////////

void MTPC::run_generator_thread()
{
    // create data and place in _main_queue; also tranfer data;
    // data is created by picking up the best next task (based
    // on already created and running create_tasks and their priority);
    // triggered via start_as_generator -> generator_thread_start
    MONITOR mon(&global_mutex); // not needed if only one generator scenario;
    while (true) {
        SIGNAL_COND signal_cond(s_global_signal_timer);
        // note that I would need such a controller for any 
        // additional generator if multi-generators-multi-workers
        // framework was used;
        MTPC::CONTROLLER::GENERATOR_DO next_do = _controller->this_generator_do();
        if (next_do == MTPC::CONTROLLER::CREATE) {
            // next function is key; it creates new a data instance, which is added
            // to main queue; data is generated from compatible, highest priority
            // task from among those in state initial;
            DATA *data = run_generator_thread_one(); // <--- key1
            if (data) {
                _controller->generator_added_task();
                _main_queue.insert_node(data);
            } else {
                data = _controller->generator_accepted_transfer();
                if (data) {
                    run_transfer_thread_one(data); // <--- key2
                    _controller->generator_finished_transfer();
                    signal_cond.signal();
                } else {
                    signal_cond.wait(&global_mutex);
                }
            }
        } else if (next_do == MTPC::CONTROLLER::TRANSFER) {
            // when workers finish work on data objects they place the corresponding tasks
            // in the "transferable" vector; from where the generator thread takes them
            // and transfers them back to the application; at the same time, the generator 
            // thread changes the status of the task to "finished";
            DATA *data = _controller->generator_accepted_transfer(); // _task_count_transferable != 0?
            assert(data);
            // next one calls app's transfer_partition_back(task) and also deletes 
            // data object;
            run_transfer_thread_one(data); // <--- key2
            _controller->generator_finished_transfer();
        } else if (next_do == MTPC::CONTROLLER::WAIT) {
            signal_cond.wait(&global_mutex);
        } else if (next_do == MTPC::CONTROLLER::END) {
            break;
        }
    }
}

MTPC::DATA *MTPC::run_generator_thread_one()
{
    while (true) {
        SIGNAL_COND signal_cond(s_global_signal_timer);
        bool has_remaining_tasks = false;
        TASK *best_next_task = 0;
        int i = 0, j = 0, best_next_i = 0;
        vector<TASK *> existing_tasks;
        existing_tasks.resize(INITIAL_VECTOR_SIZE, 0);
        int existing_tasks_count = 0;
        // () get all atsks created or running;
        for (i = 0; i < _tasks_count; i++) {
            TASK *task = get_task(i);
            switch (task->state()) {
            case TASK::RUNNING:
            case TASK::CREATED:
                existing_tasks[ existing_tasks_count ++] = task;
                break;
            default:
                break;
            }
        }
        // () look at all initial state tasks and compare for compatibility
        // with the list above; from among all compatible, select the highest
        // priority one; thus, we have created tasks with enough info to asses 
        // compatibility and set their priority;
        for (i = 0; i < _tasks_count; i++) {
            TASK *task = _tasks[i];
            if (task->state() != TASK::INITIAL)
                continue;
            has_remaining_tasks = true;
            bool compatible = true;
            for (j = 0; j < existing_tasks_count; j++) {
                if ( !_application->are_compatible(task, existing_tasks[j])) {
                    compatible = false; // e.g., overlaping or touching RECT's;
                    break;
                }
            }
            if ( !compatible)
                continue;
            if (!best_next_task) {
                best_next_task = task;
                best_next_i = i;
            } else if (task->priority() > best_next_task->priority()) {
                best_next_task = task;
                best_next_i = i;
            }
        }
        // () once, best next task is identified, we create the DATA object that
        // basically wraps the TASK and then we return the data object, which
        // will be inserted into the _main_queue;
        if (best_next_task) {
            TASK *task = best_next_task;
            // Note: in more complex applications, we may need to spend time here
            // on actually creating stuff, e.g., creating the MMEEP_WRAPPER object, 
            // populating it with info, etc.; so, here we would record how long
            // we spend on such activities;
            clock_t start_time = clock();
            // Note: should I move here the actual complete creation of the MMEEP_WRAPPER
            // object? which now I do in create_initial_tasks() of 
            // APPLICATION_SIMULATED_ANNEALING_DSE as well as in;
            MMEEP_WRAPPER *mmeep_wrapper = task->mmeep_wrapper(); 
            clock_t finish_time = clock();
            double task_data_generation_time =
                (double)(finish_time - start_time) / CLOCKS_PER_SEC;
            printf("\nGeneration time for Data of Task %d is: %.2f", 
                task->id(), task_data_generation_time);
            task->set_state(TASK::CREATED);
            DATA *data = new DATA(task, mmeep_wrapper); 
            return data;
        } else if (!has_remaining_tasks) {
            return 0;
        }
    }
}


void MTPC::run_transfer_thread_one(DATA *data)
{
    // called from run_generator_thread(); data comes from a task from the
    // transferables list, which contains tasks for which work was completed 
    // by some worker thread;
    TASK *task = data->task();

    clock_t start_time = clock();

    // here, we transfer stuff back to the application; for example, we could
    // get the final result out of the mmeep_wrapper and copy it to some place within
    // the app object;
    _application->transfer_stuff_from_task_back_to_app(task);
    task->set_state( TASK::FINISHED);

    clock_t finish_time = clock();

    double task_transfer_back_time = (double)(finish_time - start_time) / CLOCKS_PER_SEC;
    printf("\n     Task %d transfer-back time: %.2f", task->id(), task_transfer_back_time);


    // delete mmeep wrapper from the finished task, not the task;
    // this is done to free memory, which seems to explode when MEEP is launched;
    //delete task->mmeep_wrapper();
    task->call_dtor_of_mmeep_wrapper();
    

    // Note: this deletes the data not the task;
    delete data;
}


////////////////////////////////////////////////////////////////////////////////
//
// functions related to workers
//
////////////////////////////////////////////////////////////////////////////////

void MTPC::run_worker_thread()
{
    // function "passed" to worker pthread during its creation; 
    // retreive a data from _main_queue and process;
    DATA *data = static_cast<DATA *>(_main_queue.get_node());
    while ( data) {
        run_worker_thread_one( data);
        data = static_cast<DATA *>(_main_queue.get_node());
    }
}

void MTPC::run_worker_thread_one(DATA *data)
{
    // function called by the function passed to pthread during its creation; 
    // this is essentially what we want a worker thread to execute on one 
    // data instance;
    MMEEP_WRAPPER *mmeep_wrapper = data->mmeep_wrapper();
    TASK *task = data->task();
    task->set_state( TASK::RUNNING);

    _controller->worker_accepted_task();

    clock_t start_time = clock();

    // () here is really where the main work is done by a worker thhread;
    // in our case, we want here to run a meep simulation;
    // Note: "result" of this meep simulation is saved inside the mmeep_wrapper
    // object;

    mmeep_wrapper->do_meep_simulation_to_evaluate_solution();
    

    clock_t finish_time = clock();
    double task_processing_time = (double)(finish_time - start_time) / CLOCKS_PER_SEC;
    printf("\n   Task %d processing time: %.2f", task->id(), task_processing_time);

    _controller->worker_finished_task(data); // place data in transferable_tasks;
    s_global_signal_timer.signal();

}


////////////////////////////////////////////////////////////////////////////////
//
// MTPC::CONTROLLER
//
////////////////////////////////////////////////////////////////////////////////

void MTPC::CONTROLLER::print_controller_status()
{
    MONITOR mon(&_m);
    printf("\n --- State of controller of mtpc application");
    printf("\n Total tasks %d", _task_count);
    printf("\n Finished %d, Transfering %d Transferable %d Being processed %d",
           _task_count_finished, _task_count_transfering,
           _task_count_transferable, (_worker_count - _worker_count_avail));
}
void MTPC::CONTROLLER::initialize(int task_count, int worker_count)
{
    MONITOR mon(&_m);
    _task_count = task_count;
    _task_count_generated = 0;
    _task_count_avail = 0;
    _worker_count = worker_count;
    _worker_count_avail = worker_count;
    _task_count_transferable = 0;
    _task_count_transfering = 0;
    _task_count_finished = 0;
}
void MTPC::CONTROLLER::generator_added_task()
{
    MONITOR mon(&_m);
    _task_count_avail ++;
    _task_count_generated ++;
}
MTPC::DATA *MTPC::CONTROLLER::generator_accepted_transfer()
{
    MONITOR mon(&_m);
    if ( _task_count_transferable == 0) return 0;
    _task_count_transfering ++;
    _task_count_transferable --; // first decrement and then retreive data;
    DATA *data = _transferable_tasks[ _task_count_transferable];
    return data;
}
void MTPC::CONTROLLER::generator_finished_transfer()
{
    MONITOR mon(&_m);
    _task_count_finished ++;
    _task_count_transfering --;
}
void MTPC::CONTROLLER::worker_accepted_task()
{
    MONITOR mon(&_m);
    _task_count_avail --;
    _worker_count_avail --;
}
void MTPC::CONTROLLER::worker_finished_task(DATA *data)
{
    MONITOR mon(&_m);
    assert(data);
    _transferable_tasks[ _task_count_transferable] = data;
    _task_count_transferable ++;
    _worker_count_avail ++;
}
MTPC::CONTROLLER::GENERATOR_DO MTPC::CONTROLLER::this_generator_do()
{
    MONITOR mon(&_m);
    if ( (_task_count_generated < _task_count) &&
         (_task_count_avail < _worker_count_avail ||
          (_worker_count_avail == 0 && _task_count_avail <= _worker_count + 1)) ) {
        return CREATE;
    }
    if (_task_count_transferable > 0)
        return TRANSFER; // try transfering if cannot generate;
    if (_task_count_finished == _task_count) return END;
    return WAIT; // goto sleep;
}

void MTPC::CONTROLLER::incr_task_count() 
{
    // when new tasks are added to the vector of tasks, we need to increment
    // this here for the controller to be aware that tehre exist more tasks
    // than the initial number created during the initial tasks creation;
    MONITOR mon(&_m);
    _task_count ++;
}


////////////////////////////////////////////////////////////////////////////////
//
// APPLICATION_SIMULATED_ANNEALING_DSE
//
////////////////////////////////////////////////////////////////////////////////


APPLICATION_SIMULATED_ANNEALING_DSE::APPLICATION_SIMULATED_ANNEALING_DSE() 
{
    _implementation_type = MULTITHREADED; // default;
    _param_count = 4; // default;

    // by default we do a DSE by exploring only these many solutions;
    _max_num_of_tasks = 200; // default 100;
    _history_cost.resize( _max_num_of_tasks, 0.0);
    _history_cost_c1.resize( _max_num_of_tasks, 0.0);
    _history_cost_c2.resize( _max_num_of_tasks, 0.0);

    // TODO: replace nfreq = 101 with a variable provided by user?
    _nfreq = 101;
    _best_frequency.resize( 101, 0.0);
    _best_S11.resize( 101, 0.0);

    // this keeps track of how many tasks have been created so far and added 
    // to the queue;
    _total_tasks_so_far = 0;
    // annealing variables;
    _T_max = 1.0;
    _T_current = 1.0;
    _delta_T = 1.0 / double(_max_num_of_tasks); // 0.01;
    printf("\n --- delta_T = %.4f", _delta_T);

    // best solution so far initialization;
    // something very large to be sure gets overwritten by first evaluated design solution;
    _b_cost = 1000000.0; 
    
    _c_cost = 1000000.0;

    // random number generation;
    long actual_seed = time(NULL); // time(NULL) or 1 for debug; 
    _rng.set_seed( actual_seed);
}


int APPLICATION_SIMULATED_ANNEALING_DSE::create_initial_tasks()
{
    // create initially a number of tasks equal to the number of cores;
    // they will be worked on by the worker threads;
    int created_tasks_count = 0;

    // some sketch variables used for creating mmeep wrapper instances;
    double this_patch_length = 4.0;
    double this_patch_width = 4.9;
    double this_recess = 1.0;
    double this_depth = 0.2;
    int this_mmeep_wrapper_id = -1;

    // create here tasks and add them to the queue of the mtpc object;
    for ( int k = 0; k < mtpc()->cpus_count(); k ++) {

        // () we use for now a priority that is the same as the order of creation;
        int priority = _total_tasks_so_far;
        char buf[255];
        sprintf(buf, "{ task with priority %d }", priority);
        string name_s(buf);

        // () create new mmeep wrapper to pass it to new task shortly;
        // we need to generate randomly tuples of 4 different values for the parameters
        // we sweep;
        for ( int i = 0; i < _param_count; i ++) {
            // for each of the parameters, pick a "j" index randomly; between 0 and
            // max number of values for this "i" parameter;
            int j = 0; // if just one value for this param, we'll just use it;
            if ( param_values_counts(i) > 1) { // if more values, pick one randomly;
                j = _rng.flat_l(param_values_counts(i));
            }
            if ( i == 0) {
                this_patch_length = param_values(i,j);
            } else if (i == 1) {
                this_patch_width = param_values(i,j);
            } else if (i == 2) {
                this_recess = param_values(i,j);
            } else { // i == 3
                this_depth = param_values(i,j);
            }
        }
        // mmeep wrapper id is the same as the index of the task;
        this_mmeep_wrapper_id = _total_tasks_so_far;
        // create "this" mmeep_wrapper;
        MMEEP_WRAPPER *this_mmeep_wrapper = new MMEEP_WRAPPER( this_mmeep_wrapper_id,
            this_patch_length,
            this_patch_width,
            this_recess,
            this_depth);

        // () create new task and add it to the vector of tasks inside MTPC;
        // Note: that state of task is set as INITIAL;
        MTPC::TASK *task = mtpc()->add_task(this_mmeep_wrapper, priority, name_s);
        printf("\nAdded Task %d %s",
               task->id(), mtpc()->get_task(_total_tasks_so_far)->name().c_str());

        // () increment general counter to keep track how many tasks we created so far;
        _total_tasks_so_far ++;
        // increment local counter too;
        created_tasks_count ++;
    }    

    return created_tasks_count;
}


void APPLICATION_SIMULATED_ANNEALING_DSE::transfer_stuff_from_task_back_to_app(MTPC::TASK *task)
{
    MONITOR mon(&mtpc()->lock());

    MMEEP_WRAPPER *finished_mmeep_wrapper = task->mmeep_wrapper();



    // () irrespective of whether we continue to generate solution points or not,
    // we must record the cost of this one that was just finished by some worker
    // thread and which we are now "transferring" back to application;
    _history_cost[ finished_mmeep_wrapper->get_id() ] = finished_mmeep_wrapper->cost();
    _history_cost_c1[ finished_mmeep_wrapper->get_id() ] = finished_mmeep_wrapper->cost_c1();
    _history_cost_c2[ finished_mmeep_wrapper->get_id() ] = finished_mmeep_wrapper->cost_c2();
    


    // () one time deal only, we record the "current" solution, which is always used
    // as basis for generating new, child solutions;
    if ( _c_cost == 1000000.0) { // set initially in Ctor;
        _c_patch_length = finished_mmeep_wrapper->patch_length();
        _c_patch_width = finished_mmeep_wrapper->patch_width();
        _c_recess = finished_mmeep_wrapper->recess();
        _c_depth = finished_mmeep_wrapper->depth();
        _c_cost = finished_mmeep_wrapper->cost();
    }
    


    // () here is basically the logic of the annealing algorithm
    // described in the pseudocode;

    calculate_current_temperature(); // update _T_current;
    printf("\n --- T_current = %.2f", _T_current);

    if ( finished_mmeep_wrapper->cost() <= _c_cost) {
        // ---> downhill move, accept it as current;
        _c_patch_length = finished_mmeep_wrapper->patch_length();
        _c_patch_width = finished_mmeep_wrapper->patch_width();
        _c_recess = finished_mmeep_wrapper->recess();
        _c_depth = finished_mmeep_wrapper->depth();
        _c_cost = finished_mmeep_wrapper->cost();

        // need to record as best also?
        if ( finished_mmeep_wrapper->cost() <= _b_cost) {
            _b_patch_length = finished_mmeep_wrapper->patch_length();
            _b_patch_width = finished_mmeep_wrapper->patch_width();
            _b_recess = finished_mmeep_wrapper->recess();
            _b_depth = finished_mmeep_wrapper->depth();
            _b_cost = finished_mmeep_wrapper->cost();
            _b_cost_c1 = finished_mmeep_wrapper->cost_c1();
            _b_cost_c2 = finished_mmeep_wrapper->cost_c2();
            _best_center_freq = finished_mmeep_wrapper->center_freq();
            _best_BW = finished_mmeep_wrapper->BW();
            for ( int k = 0; k < finished_mmeep_wrapper->get_nfreq(); k ++) {
                _best_frequency[ k] = finished_mmeep_wrapper->frequency( k);
                _best_S11[ k] = finished_mmeep_wrapper->S11( k);
            }
            printf("\nBest so far L=%.4f W=%.4f r=%.4f d=%.4f CenterFreq=%.4f BW=%.4f", 
                _b_patch_length, _b_patch_width, _b_recess, _b_depth, 
                _best_center_freq, _best_BW);
        }
    }
    else {
        // ---> uphill move, accept as current "maybe";
        double r = _rng.sflat01(); // some random number in range 0 to 1;
        double prob = exp( -abs(_c_cost - finished_mmeep_wrapper->cost())/(_T_current) );
        printf("\n --- probability = %.2f", prob);
        if ( r < prob) {
            _c_patch_length = finished_mmeep_wrapper->patch_length();
            _c_patch_width = finished_mmeep_wrapper->patch_width();
            _c_recess = finished_mmeep_wrapper->recess();
            _c_depth = finished_mmeep_wrapper->depth();
            _c_cost = finished_mmeep_wrapper->cost();
        }
    }



    // () find out what the index of each parameter in "current" solution is;
    // TODO: this should be dome more elegantly;
    int index = 0;
    for (int j = 0; j < param_values_counts(0); j ++) {
        if ( _c_patch_length == param_values(0,j)) {
            index = j;
            break;
        }
    }
    _index_of_current_solution[0] = index;
    index = 0;
    for (int j = 0; j < param_values_counts(1); j ++) {
        if ( _c_patch_width == param_values(1,j)) {
            index = j;
            break;
        }
    }
    _index_of_current_solution[1] = index; 
    index = 0;
    for (int j = 0; j < param_values_counts(2); j ++) {
        if ( _c_recess == param_values(2,j)) {
            index = j;
            break;
        }
    }
    _index_of_current_solution[2] = index;
    index = 0;
    for (int j = 0; j < param_values_counts(3); j ++) {
        if ( _c_depth == param_values(3,j)) {
            index = j;
            break;
        }
    }
    _index_of_current_solution[3] = index;         



    // () if the total number of tasks created so far reached the maximum allowed,
    // then, we stop creating new tasks;
    if ( _total_tasks_so_far >= _max_num_of_tasks) {
        return;
    }


    // () if gotten here, we continue to create tasks and pass them via the generator thread
    // to the vector of tasks, which will be continued to be passed to worker threads, 
    // etc.;
    


    // () generate new task;
    // this is similar to how we did it inside create_initial_tasks();


    // () we use for now a priority that is the same as the order of creation;
    int priority = _total_tasks_so_far;
    char buf[255];
    sprintf(buf, "{ task with priority %d }", priority);
    string name_s(buf);

    // some sketch variables used for creating mmeep wrapper instances;
    double this_patch_length = _c_patch_length; // i = 0
    double this_patch_width = _c_patch_width; // i = 1
    double this_recess = _c_recess; // i = 2
    double this_depth = _c_depth; // i = 3
    int this_mmeep_wrapper_id = _total_tasks_so_far;
    // pick up randomly one parameter between 0 .. _param_count-1;
    // during the first half of solutions generated we allow any other 
    // value for the picked-up parameter; during the lasthalf, we allow
    // omly values from close vicinities;
    int half_max_num_of_tasks = _max_num_of_tasks / 2;
    int i = _rng.flat_l(_param_count);
    //printf("\n i = %d",i);
    int j = 0;
    if ( i == 0) {
        // pick up randomly a different value for this "i" parameter;
        // but make sure it's a different one then the existing solution
        // that we use to generate a new one from, through this SA "move";
        if ( _total_tasks_so_far <= half_max_num_of_tasks) {
            j = _rng.flat_l(param_values_counts(i));
            while ( this_patch_length == param_values(i,j)) {
                j = _rng.flat_l(param_values_counts(i)); 
            }
        } else {
            double local_r = _rng.sflat01();
            //printf("\n ---> local_r %.2f", local_r);
            if ( local_r < 0.5) {
                // pick up left index parameter value;
                j = (_index_of_current_solution[i] > 0) ? 
                    (_index_of_current_solution[i] - 1) : (_index_of_current_solution[i] + 2);
            } else { // local_r >= 0.5
                j = (_index_of_current_solution[i] < (param_values_counts(i)-1)) ? 
                    (_index_of_current_solution[i] + 1) : (_index_of_current_solution[i] - 2);
            }
        }
        this_patch_length = param_values(i,j);
    } else if (i == 1) {
        if ( _total_tasks_so_far <= half_max_num_of_tasks) {
            j = _rng.flat_l(param_values_counts(i)); 
            while ( this_patch_width == param_values(i,j)) {
                j = _rng.flat_l(param_values_counts(i)); 
            }
        } else {
            double local_r = _rng.sflat01();
            //printf("\n ---> local_r %.2f", local_r);
            if ( local_r < 0.5) {
                // pick up left index parameter value;
                j = (_index_of_current_solution[i] > 0) ? 
                    (_index_of_current_solution[i] - 1) : (_index_of_current_solution[i] + 2);
            } else { // local_r >= 0.5
                j = (_index_of_current_solution[i] < (param_values_counts(i)-1)) ? 
                    (_index_of_current_solution[i] + 1) : (_index_of_current_solution[i] - 2);
            }
        }
        this_patch_width = param_values(i,j);
    } else if (i == 2) {
        if ( _total_tasks_so_far <= half_max_num_of_tasks) {
            j = _rng.flat_l(param_values_counts(i)); 
            while ( this_recess == param_values(i,j)) {
                j = _rng.flat_l(param_values_counts(i)); 
            }
        } else {
            double local_r = _rng.sflat01();
            //printf("\n ---> local_r %.2f", local_r);
            if ( local_r < 0.5) {
                // pick up left index parameter value;
                j = (_index_of_current_solution[i] > 0) ? 
                    (_index_of_current_solution[i] - 1) : (_index_of_current_solution[i] + 2);
            } else { // local_r >= 0.5
                j = (_index_of_current_solution[i] < (param_values_counts(i)-1)) ? 
                    (_index_of_current_solution[i] + 1) : (_index_of_current_solution[i] - 2);
            }
        }
        this_recess = param_values(i,j);
    } else { // i == 3
        if ( _total_tasks_so_far <= half_max_num_of_tasks) {
            j = _rng.flat_l(param_values_counts(i)); 
            while ( this_depth == param_values(i,j)) {
                j = _rng.flat_l(param_values_counts(i)); 
            }
        } else {
            double local_r = _rng.sflat01();
            //printf("\n ---> local_r %.2f", local_r);
            if ( local_r < 0.5) {
                // pick up left index parameter value;
                j = (_index_of_current_solution[i] > 0) ? 
                    (_index_of_current_solution[i] - 1) : (_index_of_current_solution[i] + 2);
            } else { // local_r >= 0.5
                j = (_index_of_current_solution[i] < (param_values_counts(i)-1)) ? 
                    (_index_of_current_solution[i] + 1) : (_index_of_current_solution[i] - 2);
            }
        }
        this_depth = param_values(i,j);
    }


    // () create a new mmeep_wrapper;
    MMEEP_WRAPPER *this_mmeep_wrapper = new MMEEP_WRAPPER( this_mmeep_wrapper_id,
        this_patch_length, 
        this_patch_width,
        this_recess,
        this_depth);


    // () create new task and add it to the vector of tasks inside MTPC;
    // Note: that state of task is set as INITIAL;
    MTPC::TASK *this_task = mtpc()->add_task(this_mmeep_wrapper, priority, name_s);
    printf("\nAdded Task %d %s",
           this_task->id(), mtpc()->get_task(_total_tasks_so_far)->name().c_str());


    // () increment general counter to keep track how many tasks we created so far;
    _total_tasks_so_far ++;


    // also, reflect that we added a task in the controller too, so that generator
    // thread will continue to add the tasks to the queue;
    mtpc()->get_controller()->incr_task_count() ;
    

    return;
}


bool APPLICATION_SIMULATED_ANNEALING_DSE::are_compatible(MTPC::TASK *t1, MTPC::TASK *t2)
{
    // empty, because we do not use the concept of compatibility here;
    return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// functions related to parameters from user
//
////////////////////////////////////////////////////////////////////////////////


// we'll give as input a file with params; 
// to make my life easy, the format of this input file is expected to be 
// like this:
//
// 4                 <--- number of parameters to sweep
// patch_length 3    <--- patch length [cm] and number of values on next line
// 3.5 4.0 4.5       <--- the discrete values of the "length" parameter
// patch_width 2     <--- patch width [cm] and number of values on next line
// 4.9 5.0           <--- the discrete values of the "width" parameter
// recess_distance 2 <--- recess distance r [cm] and number of values on next line
// 1 1.2             <--- values
// gap distance 3    <--- gap distance d [cm] and number of values on next line
// 0.4 0.5 0.6       <--- values
//
// Note: values for a parameter do not need to be given in any specific order;


bool APPLICATION_SIMULATED_ANNEALING_DSE::read_parameters_from_file()
{
    // read parameters from input file provided by user;


    // (1) open file;
    ifstream params_ifstream;
    params_ifstream.open( _in_file_name.c_str());
    if ( !params_ifstream) {
        printf("\nError: Cannot open parameters file: <%s>\n", _in_file_name.c_str());
        exit(1);
    } else {
        printf("\nOpened parameters file <%s> for reading...", _in_file_name.c_str());
    }
    

    // (2) read lines from file; process as per our agreement;
    string line;

    // 1st line must contain an integer;
    if ( getline( params_ifstream, line)) {
        istringstream iss( line);
        string tmp_string;
        iss >> _param_count;
        
        // size-up all vectors;
        _param_names.resize( _param_count);
        _param_values_counts.resize( _param_count);
        _index_of_current_solution.resize( _param_count);
        _param_values.resize( _param_count); // individual "i" will be sized-up later;
    } else {
        printf("\nError: Reading from file issue: <%s>\n", _in_file_name.c_str()); exit(1);
    }
    

    // next lines are in groups of two for each parameter;
    for ( int i = 0; i < _param_count; i ++) {
        // one line with name of parameter and number of values;
        if ( getline( params_ifstream, line)) {
            istringstream iss( line);
            string tmp_string;
            iss >> tmp_string; // name;
            _param_names[i] = tmp_string;
            int tmp_int;
            iss >> tmp_int; // count;
            _param_values_counts[i] = tmp_int;
            _param_values[i].resize( _param_values_counts[i]); 
        } else {
            printf("\nError: Reading from file issue: <%s>\n", _in_file_name.c_str()); exit(1);
        }
        // another line with values;
        if ( getline( params_ifstream, line)) {
            istringstream iss( line);
            for ( int j = 0; j < param_values_counts(i); j ++) {
                double tmp_double;
                iss >> tmp_double;
                _param_values[i][j] = tmp_double;
            }
        } else {
            printf("\nError: Reading from file issue: <%s>\n", _in_file_name.c_str()); exit(1);
        }
    }


    // (3) close input file;
    params_ifstream.close();
    printf("\nDone reading file <%s>.\n", _in_file_name.c_str());

    return true;
}


void APPLICATION_SIMULATED_ANNEALING_DSE::print_parameters()
{
    // this prints params we read from input file;
    // each param has a list of possible values;

    printf("\nListing of parameters from file <%s>:", _in_file_name.c_str());
    printf("\nParameters count: %d", _param_count);

    for ( int i = 0; i < _param_count; i ++) {
        printf("\n(%d) %s  --  num_values: %d\n", i, param_name(i).c_str(), param_values_counts(i));
        for ( int j = 0; j < param_values_counts(i); j ++) {
            printf("%.2f  ", param_values(i,j));
        }
    }

    printf("\n\nSimulations will be done using implementation type: ");
    if ( _implementation_type == SEQUENTIAL) {
        printf("SEQUENTIAL");
    } else if (_implementation_type == MULTITHREADED) {
        printf("MULTITHREADED");
    } else {
        printf("UNKNOWN (ERROR)");
    }

    printf("\n");
}

void APPLICATION_SIMULATED_ANNEALING_DSE::print_best_found_solution()
{

    printf("\n\nBEST DESIGN SOLUTION:\n================\n");

    printf("\nBest patch length: %.4f", _b_patch_length);
    printf("\nBest patch width: %.4f", _b_patch_width);
    printf("\nBest patch recess: %.4f", _b_recess);
    printf("\nBest patch depth: %.4f", _b_depth);

    for ( int i = 0; i < _nfreq; i ++) {
        printf("\n%d \t %.2f \t %.2f", i, _best_frequency[i], _best_S11[i]);
    }

    printf("\n\nBest Center Frequency = %.2f Bandwidth = %.2f", _best_center_freq, _best_BW);
    
    printf("\n\nSIMULATED ANNEALING:\n================\n");

    printf("\nCost1 \t Cost2 \t Cost");
    for ( int j = 0; j < _max_num_of_tasks; j ++) {
        printf("\n%.2f \t %.2f \t %.2f", _history_cost_c1[j], _history_cost_c2[j],
            _history_cost[j]);
    }
}

////////////////////////////////////////////////////////////////////////////////
//
// RANDOM_NUMBER_GENERATOR
//
////////////////////////////////////////////////////////////////////////////////

RANDOM_NUMBER_GENERATOR::RANDOM_NUMBER_GENERATOR(long seed) :_seed(seed)
{
    srandom( _seed);
}

RANDOM_NUMBER_GENERATOR::RANDOM_NUMBER_GENERATOR() : _seed(1)
{
    srandom( _seed);
}

double RANDOM_NUMBER_GENERATOR::sflat01()
{
    double val = random() * 1.0 / RAND_MAX;
    return val;
}

void RANDOM_NUMBER_GENERATOR::set_seed(long seed) 
{
    _seed = seed;
    srandom( seed);
}

double RANDOM_NUMBER_GENERATOR::flat_d(double low, double high) 
{
    assert( low < high);
    return (high - low) * sflat01() + low;
}

long RANDOM_NUMBER_GENERATOR::flat_l(long high_plus_1)
{
    // generate a random number in the range 0 to "high";
    // note: use when high_plus_1 is small;
    long val = random() % high_plus_1;
    assert(val >= 0 && val < high_plus_1);
    return val;
}

long RANDOM_NUMBER_GENERATOR::flat_l_low_high(long low, long high)
{
    // note: use when the interval low-high is relatively large;
    assert( low < high);
    long val = long((high - low) * sflat01() + low);
    assert( val >= low && val <= high);
    return val;
}

unsigned long RANDOM_NUMBER_GENERATOR::flat_ul(unsigned long low, unsigned long high) 
{
    assert( low < high);
    unsigned long val = (unsigned long)((high-low) * sflat01() + low);
    assert( val >= low && val <= high);
    return val;
}

unsigned long long RANDOM_NUMBER_GENERATOR::flat_ull(unsigned long long low,
    unsigned long long high) 
{
    assert( low < high);
    unsigned long long val = (unsigned long long)((high-low) * sflat01() + low);
    assert( val >= low && val <= high);
    return val;
}

double RANDOM_NUMBER_GENERATOR::gauss01()
{
    // mean = 0, variance = 1;
    double modifier;
    double compile_b;
    double in_a, in_b;
    double out_a;
    static double out_b;
    static int recalc = 1;

    if (recalc) {
        // Range from (0:1], not [0:1). Had to change this to prevent log(0).
        in_a = 1.0 - sflat01();
        in_b = sflat01();

        modifier = sqrt( -2.0 * log(in_a));
        compile_b = 2.0 * PI * in_b;

        out_a = modifier * cos(compile_b);
        out_b = modifier * sin(compile_b);

        recalc = 0;
        return out_b;
    }

    recalc = 1;
    return out_b;
}

double RANDOM_NUMBER_GENERATOR::gauss_mean_d(double mean, double variance) 
{
    double ret = gauss01() * variance + mean;
    return ret;
}

long RANDOM_NUMBER_GENERATOR::gauss_mean_l(long mean, double variance) 
{
    return ((long)(gauss01() * variance + mean));
}

unsigned long RANDOM_NUMBER_GENERATOR::gauss_mean_ul(unsigned long mean, 
    double variance) 
{
    return((unsigned long)(gauss01() * variance + mean));
}

unsigned long long RANDOM_NUMBER_GENERATOR::gauss_mean_ull(unsigned long long mean, 
    double variance) 
{
    return ((unsigned long long)(gauss01() * variance + mean));
}


