#ifndef __MTPC_APP__
#define __MTPC_APP__

#include <stdio.h>
#include <stdlib.h>
#include <vector>

#include "mtpc_basics.h"
#include "mmeep_wrapper.h"


#define INITIAL_VECTOR_SIZE 128 // allocate initialy to avoid often mem alloc; 

class APPLICATION_SIMULATED_ANNEALING_DSE;


////////////////////////////////////////////////////////////////////////////////
//
// some utilities;
//
////////////////////////////////////////////////////////////////////////////////

static int my_rand_int(int imax) 
{
    // return a random integer between [0 .. imax-1]
    int ival = rand() % imax;
    return ival;
}


////////////////////////////////////////////////////////////////////////////////
//
// RANDOM_NUMBER_GENERATOR
//
////////////////////////////////////////////////////////////////////////////////

const double PI = 3.141592658979323846;

class RANDOM_NUMBER_GENERATOR
{
 private:
    long _seed;
 public:
    double sflat01();
    double gauss01();

 public:
    RANDOM_NUMBER_GENERATOR(long seed);
    RANDOM_NUMBER_GENERATOR();
    ~RANDOM_NUMBER_GENERATOR() {}

    double flat_d( double low, double high);
    long flat_l_low_high( long low, long high);
    long flat_l( long high_plus_1);
    unsigned long flat_ul( unsigned long low, unsigned long high);
    unsigned long long flat_ull( unsigned long long low, unsigned long long high);
    double gauss_mean_d( double mean, double variance);
    long gauss_mean_l( long mean, double variance);
    unsigned long gauss_mean_ul( unsigned long mean, double variance);
    unsigned long long gauss_mean_ull( unsigned long long mean, double variance);
    void set_seed( long seed);
};


////////////////////////////////////////////////////////////////////////////////
//
// MTPC
//
////////////////////////////////////////////////////////////////////////////////

// MTPC: Multi-Threaded Platform in C/C++
// provides the platform on which we run multithreaded "applications",
// using the generators-workers framework described in mtpc_basics.h;

class MTPC : public THREADED_FRAMEWORK {
 public:
    class CONTROLLER;
    class TASK;
    class DATA;
 private:
    APPLICATION_SIMULATED_ANNEALING_DSE *_application;
    CONTROLLER *_controller; // takes care of part of synchronization;
    vector<TASK *> _tasks;
    int _tasks_count;
    int _cpus_count; // same as maintained by general director;
    MUTEX _lock;
    static COND _task_cond;
 protected:
    void run_generator_thread();
    void run_worker_thread();
    DATA *run_generator_thread_one();
    void run_worker_thread_one(DATA *data);
    void run_transfer_thread_one(DATA *data);

 public:
    MTPC(APPLICATION_SIMULATED_ANNEALING_DSE *application);
    ~MTPC();
    
    int cpus_count() const { return _cpus_count; }
    MUTEX &lock() { return _lock; }
    CONTROLLER *get_controller() { return _controller; }
    TASK *add_task(MMEEP_WRAPPER *mmeep_wrapper, int priority, string &name);
    TASK *get_task(int i) const {
        assert( i >= 0 && i < _tasks_count);
        return _tasks[i];
    }

    int run();
    
    friend class APPLICATION_SIMULATED_ANNEALING_DSE;
};

////////////////////////////////////////////////////////////////////////////////
//
// MTPC::CONTROLLER
//
////////////////////////////////////////////////////////////////////////////////

// a stand alone class in here used to facillitate sync;

class MTPC::CONTROLLER { 
 private:
    MUTEX _m; // used to lock routines of this class;
    MTPC *_mtpc;
    int _task_count; // all, within an application;
    int _task_count_generated; // produced by generator(s);
    int _task_count_avail; // in queue, to be pickd-up by workers;
    int _worker_count; // all;
    int _worker_count_avail;
    int _task_count_transferable;
    int _task_count_transfering;
    int _task_count_finished;
    vector<DATA *> _transferable_tasks;
 public:
    enum GENERATOR_DO { CREATE = 0, TRANSFER = 1, END = 2, WAIT = 3 };

 public:
    CONTROLLER(MTPC *mtpc) : _m("CONTROLLER_class_lock"), _mtpc(mtpc) {
        _task_count = 0;
        _task_count_generated = 0;
        _task_count_avail = 0;
        _worker_count = 0;
        _worker_count_avail = 0;
        _task_count_transferable = 0;
        _task_count_transfering = 0;
        _task_count_finished = 0;
        _transferable_tasks.resize(INITIAL_VECTOR_SIZE, 0);
    }
    ~CONTROLLER() {}

    void print_controller_status(); // debug;
    void initialize(int task_count, int worker_count);
    void generator_added_task();
    MTPC::DATA *generator_accepted_transfer();
    void generator_finished_transfer();
    GENERATOR_DO this_generator_do();
    void worker_accepted_task();
    void worker_finished_task(DATA *data);
    // needed for SA algorithm;
    void incr_task_count();
    
};

////////////////////////////////////////////////////////////////////////////////
//
// MTPC::TASK
//
////////////////////////////////////////////////////////////////////////////////

// a task is basically representing one instance of meep execution,
// the meep simulation is run from within the member MMEEP_WRAPPER; 

class MTPC::TASK {
 public:
    enum STATE { INITIAL = 0, CREATED = 1, RUNNING = 2, FINISHED = 3 };
 private:
    int _id;
    int _priority;
    string _name;
    STATE _state;
    MMEEP_WRAPPER *_mmeep_wrapper; // do we even need it here or only inside DATA?

 public:
    TASK(MMEEP_WRAPPER *mmeep_wrapper, int priority, string &name) : 
        _id(-1), _priority(priority), _name(name), _state(INITIAL), 
        _mmeep_wrapper(mmeep_wrapper) {}
    ~TASK() {}

    MMEEP_WRAPPER *mmeep_wrapper() { return _mmeep_wrapper; }
    void call_dtor_of_mmeep_wrapper() { _mmeep_wrapper->~MMEEP_WRAPPER(); }
    STATE state() const { return _state; }
    void set_state(STATE state) {
        assert(state > _state);
        _state = state;
    }
    int priority() const { return _priority; }
    const string &name() const { return _name; }
    int id() const { return _id; }
    void set_id(int id) { _id = id; }
};

////////////////////////////////////////////////////////////////////////////////
//
// MTPC::DATA
//
////////////////////////////////////////////////////////////////////////////////

class MTPC::DATA : public MTPC_NODE {
    TASK *_task;
    MMEEP_WRAPPER *_mmeep_wrapper;

 public:
    DATA(TASK *task, MMEEP_WRAPPER *mmeep_wrapper) : _task(task), _mmeep_wrapper(mmeep_wrapper) {}
    ~DATA() {}

    TASK *task() const { return _task; }
    MMEEP_WRAPPER *mmeep_wrapper() const { return _mmeep_wrapper; }
};


////////////////////////////////////////////////////////////////////////////////
//
// APPLICATION_SIMULATED_ANNEALING_DSE
//
////////////////////////////////////////////////////////////////////////////////

enum IMPLEMENTATION_TYPE { SEQUENTIAL=0, MULTITHREADED=1 };

// simulated annealing (SA) based design space exploration (DSE) for
// patch antennas;


class APPLICATION_SIMULATED_ANNEALING_DSE {
 private:
    MTPC *_mtpc;
    int _total_tasks_so_far;

    RANDOM_NUMBER_GENERATOR _rng;
    long _rng_seed;

 private:
    // sequential or multithreaded?
    IMPLEMENTATION_TYPE _implementation_type;
    // total number of parameters that we'll sweep during SA;
    int _param_count; 
    // names of parameters;
    vector<string> _param_names;
    // number of different values for a given parameter;
    vector<int> _param_values_counts;
    // next one stores index of parameter that is part or "current" solution in the
    // simulated annealing alog; needed to be able to work easier with larger or smaller
    // vicinities around current parameter value during annealing;
    vector<int> _index_of_current_solution;
    // actual numerical values for each parameter;
    vector<vector<double> > _param_values;

 public:
    string _in_file_name; // parameters file, provided by user;
    string _out_file_name;

 public:
    // best, "b", solution stored here;
    double _b_patch_length; 
    double _b_patch_width; 
    double _b_recess; 
    double _b_depth; 
    double _b_cost, _b_cost_c1, _b_cost_c2;
    double _best_center_freq, _best_BW;
    int _nfreq;
    vector<double> _best_frequency; // nfreq=101 entries;
    vector<double> _best_S11; // nfreq=101 entries;
    
    // current, "c", solution stored here; used during annealing;
    double _c_patch_length; 
    double _c_patch_width; 
    double _c_recess; 
    double _c_depth; 
    double _c_cost;

    // sketch variable to store costs of all solutions evaluated during the
    // SA algorithm; the number of entries should be the same as the total number
    // of evaluated solutions untill SA algo stops;
    vector<double> _history_cost;
    vector<double> _history_cost_c1;
    vector<double> _history_cost_c2;

    // annealing variables;
    int _max_num_of_tasks; // hard coded to 100;
    double _T_max;
    double _T_current;
    double _delta_T;
    

 public:
    APPLICATION_SIMULATED_ANNEALING_DSE();
    ~APPLICATION_SIMULATED_ANNEALING_DSE() {}

    // create an initial population of tasks;
    int create_initial_tasks();

    void set_mtpc(MTPC *mtpc) { _mtpc = mtpc; }
    MTPC *mtpc() const { return _mtpc; }
    void set_rng_seed(long seed) { _rng_seed = seed; }
    long rng_seed() const { return _rng_seed; }
    RANDOM_NUMBER_GENERATOR &rng() { return _rng; }

    bool are_compatible(MTPC::TASK *t1, MTPC::TASK *t2);

    void transfer_stuff_from_task_back_to_app(MTPC::TASK *task);

    double calculate_current_temperature() {
        // should use as input arguments iteration index k, and _T_max;
        // possibly also an annealing schedule for smarter implementations;
        //_T_current = max( 0.0, _T_max - (_delta_T * _total_tasks_so_far) );
        _T_current = max( 0.001, _T_current - _delta_T );
        return _T_current;
    }
    

    // () functions related to parameter values from user;
    // sets;
    void set_implementation_type( IMPLEMENTATION_TYPE type) { _implementation_type = type; }
    void set_param_count( int param_count) { _param_count = param_count; }
    // Note: these functions should be called only after vectors have been 
    // sized up first! 
    void set_param_name( int i, string param_name) {
        // sanity check;
        assert( i >= 0 && i < _param_count);  
        _param_names[i] = param_name; 
    }
    void set_param_values_counts( int i, int this_count) {
        // sanity check;
        assert( i >= 0 && i < _param_count);  
        _param_values_counts[i] = this_count; 
    }
    void set_param_values( int i, int j, double this_value) {
        // sanity checks;
        assert( i >= 0 && i < _param_count);
        assert( j >= 0 && j < _param_values_counts[i]);    
        _param_values[i][j] = this_value; 
    }
    void set_input_filename( string name) { _in_file_name = name; }
    void set_output_filename( string name) { _out_file_name = name; }
    // gets;
    IMPLEMENTATION_TYPE implementation_type() { return _implementation_type; }
    int param_count() { return _param_count; }
    vector<string> &param_names() { return _param_names; }
    vector<int> &param_values_counts() { return _param_values_counts; }
    vector<vector<double> > &param_values() { return _param_values; }
    string param_name( int i) {
        assert( i >= 0 && i < _param_count);  
        return _param_names[i];
    }
    int param_values_counts( int i) {
        assert( i >= 0 && i < _param_count);  
        return _param_values_counts[i];
    }
    double param_values( int i, int j) {
        assert( i >= 0 && i < _param_count);
        assert( j >= 0 && j < _param_values_counts[i]);    
        return _param_values[i][j];
    }

    // read the file with parameters from user;
    bool read_parameters_from_file();
    void print_parameters(); // debug; params ranges from input file;
    
    void print_best_found_solution();

    // () list here classes that are friends;
    friend class MTPC;
};

#endif
