#ifndef _MTPC_BASICS_H_
#define _MTPC_BASICS_H_

#include <sys/time.h>
#include <assert.h>
#include <stdio.h>
#include <iostream>
#include <string>
#include <semaphore.h>
#include <pthread.h>

using namespace std;

#define MAX_NUM_THREADS 128

typedef int MTPC_THREAD_ID;
typedef pthread_t MTPC_THREAD;
typedef pthread_mutex_t MTPC_MUTEX;
#define MTPC_MUTEX_TRYLOCK(m) pthread_mutex_trylock(m)
#define MTPC_MUTEX_LOCK(m) pthread_mutex_lock(m)
#define MTPC_MUTEX_UNLOCK(m) pthread_mutex_unlock(m)
#define MTPC_MUTEX_INIT(m) { int rc = pthread_mutex_init(m, 0); assert(!rc); }
#define MTPC_MUTEX_DELETE(m) { int rc = pthread_mutex_destroy(m); assert(!rc); }

typedef sem_t MTPC_SEMAPHORE;
#define SEM_WAIT(s) if (!_rc) while (sem_wait(&s));
#define SEM_RELEASE(s) if (!_rc) sem_post(&s);
#define SEM_INIT(s, n) _rc = sem_init(&_s, 0, n)
#define SEM_DELETE(s) if (!_rc) sem_destroy(&_s)

typedef pthread_cond_t MTPC_COND;
#define COND_WAIT(c, m) pthread_cond_wait(c, m);
#define COND_INIT(c) { int rc = pthread_cond_init(c, 0); assert(!rc); }
#define COND_DELETE(c) { int rc = pthread_cond_destroy(c); assert(rc); }
#define COND_SIGNAL(c) pthread_cond_signal(c);
#define COND_BROADCAST(c) pthread_cond_broadcast(c);

#define THREAD_CREATE(thread_t, func, arg) pthread_create(&thread_t, NULL, func, arg);
#define THREAD_JOIN(thread_t) pthread_join(thread_t, NULL)


////////////////////////////////////////////////////////////////////////////////
//
// DIRECTOR
//
////////////////////////////////////////////////////////////////////////////////

// will globally orchestrate thread creation, keep track of how many
// are workers/generators, bound their number; collect also 
// info about local machine;

class DIRECTOR {
 private:
    MTPC_MUTEX _mutex; // lock;
    int _threads_count;
    int _worker_threads_count; // currently working threads;
    int _cpus_count;
 public:
    DIRECTOR();
    ~DIRECTOR();

    int get_threads_count() const { return _threads_count; }
    int get_cpus_count() const { return _cpus_count; }
    void reset_thread_count() { _threads_count = 0; }
    bool incr_threads_count(bool this_is_worker); // false means generator;
    void decr_threads_count(bool this_is_worker);
    void print_director_initial() {
        MTPC_MUTEX_LOCK(&_mutex);
        printf("\n --- General director : initialized");
        printf("\n This machine has %d CPUS", _cpus_count);
        MTPC_MUTEX_UNLOCK(&_mutex);
    }
    void print_director_status() {
        MTPC_MUTEX_LOCK(&_mutex);
        printf("\n --- General director : snapshot");
        printf("\n Multithreaded framework running with generators %d and workers %d",
               _worker_threads_count, _threads_count - _worker_threads_count);
        MTPC_MUTEX_UNLOCK(&_mutex);
    }
};

// this is a global variable; the only one;
extern DIRECTOR g_director;
#define THREADS_COUNT_FROM_DIRECTOR g_director.get_threads_count()

////////////////////////////////////////////////////////////////////////////////
//
// MUTEX
//
////////////////////////////////////////////////////////////////////////////////

class MUTEX {
 private:
    string _name;
    int _id_of_last_locking_thread;
    int _mutex_count;
    int _conflict_count;
    int _conflict_time;
    MTPC_MUTEX _mutex;

    void _lock() {
        if (MTPC_MUTEX_TRYLOCK(&_mutex) != 0) {
            double walltime_before = 0; //getWallTime();
            MTPC_MUTEX_LOCK(&_mutex);
            double walltime_after = 0;//getWallTime();
            _conflict_count++;
            int usec = int((walltime_after - walltime_before) * 1e6);
            _conflict_time += usec;
        }
        _mutex_count++;
    }
    void _unlock() { assert(MTPC_MUTEX_UNLOCK(&_mutex) == 0); }
 public:
    MUTEX(const char *name = "") { 
        _name = name;
        _id_of_last_locking_thread = -1; 
        _mutex_count = 0;
        _conflict_count = 0;
        _conflict_time = 0;
        MTPC_MUTEX_INIT(&_mutex);
    }
    ~MUTEX() {
        MTPC_MUTEX_DELETE(&_mutex);
    }
    MTPC_MUTEX *get_mutex() { return &_mutex; } 
    void lock() { if (THREADS_COUNT_FROM_DIRECTOR) _lock(); }
    void unlock() { if (THREADS_COUNT_FROM_DIRECTOR) _unlock(); }
};

////////////////////////////////////////////////////////////////////////////////
//
// SEMAPHORE
//
////////////////////////////////////////////////////////////////////////////////

class SEMAPHORE {
 protected:
    MTPC_SEMAPHORE _s;
    int _rc;
 public:
    SEMAPHORE(int n) { SEM_INIT(_s, n); }
    SEMAPHORE() { SEM_INIT(_s, 0); }
    ~SEMAPHORE() { SEM_DELETE(_s); }
    void init(int n) { SEM_INIT(_s, n); }
    void release() { SEM_RELEASE(_s); }
    void wait() { SEM_WAIT(_s); }
};

////////////////////////////////////////////////////////////////////////////////
//
// COND
//
////////////////////////////////////////////////////////////////////////////////

class COND {
 private:
    MTPC_COND _c;
 public:
    COND() { COND_INIT(&_c); }
    ~COND() {}
    void wait(MUTEX *m) { COND_WAIT(&_c, m->get_mutex()); }
    void signal() { COND_SIGNAL(&_c); }
    void broadcast() { COND_BROADCAST(&_c); }
};

////////////////////////////////////////////////////////////////////////////////
//
// MONITOR
//
////////////////////////////////////////////////////////////////////////////////

// used to lock a function so that only one thread can run the function;
// say: static MTPC_MUTEX m; MONITOR(&m); at begining of function;

class MONITOR {
 private:
    MUTEX *_m;
    MUTEX *mutex() { return _m; }
 public:
    MONITOR(MUTEX *m) : _m(m) { if ( _m) _m->lock(); }
    ~MONITOR() { if ( _m) _m->unlock(); }
};

////////////////////////////////////////////////////////////////////////////////
//
// SIGNAL_TIMER
//
////////////////////////////////////////////////////////////////////////////////

class SIGNAL_TIMER {
    // timer, incremented when some thread calls signal(); sync;
    int _time;
    MUTEX _m;
    COND _cond;
 public:
    SIGNAL_TIMER() : _time(0), _m("") {}
    ~SIGNAL_TIMER() {}
    void wait_till_after(MUTEX *m, int this_time) {
        // wait/sleep until time passes this_time; release lock in meantime;
        if (m) m->unlock();
        MONITOR mon(&_m);
        while ( _time <= this_time) {
            _cond.wait(&_m);
        }
        if (m) m->lock();
    }
    void signal() { MONITOR mon(&_m); _time++; _cond.signal(); }
    int current_time() { return _time; }
};

////////////////////////////////////////////////////////////////////////////////
//
// SIGNAL_COND
//
////////////////////////////////////////////////////////////////////////////////

class SIGNAL_COND {
    // at ctor, record the time of the SIGNAL_TIMER;
    // wait() will block until after the time advanced since initial time;
    SIGNAL_TIMER &_timer;
    int _init_time;
 public:
    SIGNAL_COND(SIGNAL_TIMER &timer) :
    _timer(timer), _init_time(timer.current_time()) {}
    ~SIGNAL_COND() {}
    void signal() { _timer.signal(); }
    void wait() { _timer.wait_till_after(0, _init_time); }
    void wait(MUTEX *m) { _timer.wait_till_after(m, _init_time); }
};

////////////////////////////////////////////////////////////////////////////////
//
// MTPC_NODE
//
////////////////////////////////////////////////////////////////////////////////

class MTPC_NODE {
 public:
    MTPC_NODE *next;
};

////////////////////////////////////////////////////////////////////////////////
//
// MTPC_QUEUE
//
////////////////////////////////////////////////////////////////////////////////

// this is a general purpose simply linked list; used to implement the
// queue with tasks created by generators and processed by workers;

class MTPC_QUEUE {
 private:
    MUTEX mutex;
    MTPC_NODE *head;
    MTPC_NODE *tail;
    int length;
    bool stop;
    SEMAPHORE *green_to_produce;
    COND green_to_consume;
 public:
    MTPC_QUEUE(int max_length = 128) : mutex("queue"), head(0), tail(0),
        length(0), stop(false), green_to_produce(new SEMAPHORE(max_length)) {}
    ~MTPC_QUEUE() {
        MTPC_NODE *node;
        mutex.lock();
        while ((node = head)) {
            head = head->next;
            delete node;
        }
        length = 0;
        mutex.unlock();
        if (green_to_produce) delete green_to_produce;
    }
    void insert_node(MTPC_NODE *new_node) {
        green_to_produce->wait();
        mutex.lock();
        if (head == 0) {
            head = tail = new_node;
            new_node->next = 0;
        } else {
            tail->next = new_node;
            tail = new_node;
            new_node->next = 0; 
        }
        length++;
        green_to_consume.signal();
        mutex.unlock();
    }
    MTPC_NODE *get_node() {
        MTPC_NODE *node;
        mutex.lock();
        while ((head == 0) && (stop == false)) {
            green_to_consume.wait(&mutex);
        }
        node = head;
        if (head) {
            head = head->next;
            if (head == 0)
                tail = 0;
            length--;
        } else {
            assert(stop);
        } 
        mutex.unlock();
        if (node)
            green_to_produce->release();
        return node;
    }
    void set_stop() {
        // chemata de T principal dupa ce toti generators au exit; setarea 
        // lui stop trighereaza ca toti workers sa exit dupa ce queue nu
        // mai contine work to do;
        mutex.lock();
        stop = true; 
        green_to_consume.broadcast();
        mutex.unlock();
    }
    void reset_queue() {
        mutex.lock(); stop = false; mutex.unlock();
    }
    int compute_length() {
        mutex.lock();
        int count = 0;
        MTPC_NODE *node = head;
        while (node) { count++; node = node->next; }
        mutex.unlock();
        return count;
    }
    int get_length() const { return length; }
};

////////////////////////////////////////////////////////////////////////////////
//
// THREADED_FRAMEWORK
//
////////////////////////////////////////////////////////////////////////////////

// base class used to facilitate implementation of multithreaded applications;
// this is just the bare minimum to manipulate a number of "generators and workers"; 
// I use one generator and multiple workers though;
// note: class MTPC will inherit (i.e., will be derived from) this one;

class THREADED_FRAMEWORK {
 protected:
    MTPC_MUTEX _mutex; // lock;
    MTPC_QUEUE _main_queue;
    int _num_generators;
    int _num_workers;
    MTPC_THREAD _generators[MAX_NUM_THREADS];
    MTPC_THREAD _workers[MAX_NUM_THREADS];

 public:
    THREADED_FRAMEWORK(int queue_size = 128) : _main_queue(queue_size), 
        _num_generators(0), _num_workers(0) {
        // allocate memory for _generators and _workers if dynamic arrays;
    }
    virtual ~THREADED_FRAMEWORK() {}

    virtual void run_generator_thread() {}
    virtual void run_worker_thread() {}
    void wait_generators(int num) {
        assert(num >= _num_generators);
        for ( int i = 0; i < _num_generators; i++) {
            THREAD_JOIN( _generators[i]);
        }
        _num_generators = 0;
    }
    void wait_workers(int num) {
        assert(num >= _num_workers);
        for ( int i = 0; i < _num_workers; i++) {
            THREAD_JOIN( _workers[i]);
        }
        _num_workers = 0;
    }
    void add_generator(MTPC_THREAD this_thread) {
        _generators[_num_generators++] = this_thread;
    }
    void add_worker(MTPC_THREAD this_thread) {
        _workers[_num_workers++] = this_thread;
    }
    void reset() {
        _main_queue.reset_queue();
    }
};

////////////////////////////////////////////////////////////////////////////////
//
// ONE_THREAD
//
////////////////////////////////////////////////////////////////////////////////

class ONE_THREAD {
 protected:
    THREADED_FRAMEWORK *_framework; // pointer to this thread's host;
    MTPC_THREAD _thread;
    MTPC_THREAD_ID _thread_id;
 protected:
    ONE_THREAD(THREADED_FRAMEWORK *framework) : _framework(framework) {}
    static void *generator_thread_start(void *arg);
    static void *worker_thread_start(void *arg);
    MTPC_THREAD_ID thread_id() { return _thread_id; }
 public:
    virtual ~ONE_THREAD() {}
    static ONE_THREAD *create_thread(THREADED_FRAMEWORK *framework) {
        return new ONE_THREAD(framework);
    }
    THREADED_FRAMEWORK *get_framework() { return _framework; }
    int start_as_generator();
    int start_as_worker();
    int set_affinity_custom(int i);
};

#endif
