#include <iostream>
#include <math.h>
#include <complex>
#include <string>
#include <cstdlib>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/types.h>

#include "mtpc_basics.h"
#include "mtpc_app.h"
#include "mmeep_wrapper.h"


using namespace std;

////////////////////////////////////////////////////////////////////////////////
//
// readme 
//
////////////////////////////////////////////////////////////////////////////////

// -- everything that jim used to do as a full simulation is now inside MMEEP_WRAPPER class,
// except the four variables that we sweep during the design space exploration, DSE;
// those four, as intervals, are stored in members of the other important class,
// APPLICATION_SIMULATED_ANNEALING_DSE, which is responsible with implementing the
// simulated annealing, SA, search;
// -- each worker thread will work on DATA which encapsulates a TASK which encapsulates
// a MMEEP_WRAPPER object; 
// -- running meep, as a simulator, is done by calling a member function of the 
// MMEEP_WRAPPER object;
// -- each MMEEP_WRAPPER object is created with four different values for the four
// different parameters that we sweep as design space exploration, DSE;
// -- these four parameters values are "picked up" from within the main variable,
// "application_dse", which is an instance of APPLICATION_SIMULATED_ANNEALING_DSE;
// 
// -- the third important class is the MTPC class; this is derived from 
// THREADED_FRAMEWORK class, which in turn basically implements a 
// "generator-workers" strategy and which has a _main_queue (MTPC_QUEUE object)
// that will store DATA objects;
// -- tasks are created and added to the vector of tasks; newly added tasks are
// added with initial state; then, later the generator thread picks tasks by compatibility
// and priority and creates data objects that are added to the queue of data objects;
// -- DATA objects are added to the main queue (inherited member from THREADED_FRAMEWORK)
// by picking up the "best next TASK" from the vector of tasks maintained directly
// inside MTPC class and creating a new DATA object (DATA object hosts/wraps-around the TASK);
// -- when new data is created from best_next_task from the vector of tasks and added
// to the main_queue, the task state is changed to "created";
// -- worker threads consume concurrently data objects from main_queue and 
// do work on them; when a certain data object is removed/consumed by a worker thread
// from the queue, the corresponding task in the vector of tasks has it status updated
// or changed to "running";

////////////////////////////////////////////////////////////////////////////////
//
// main launching point;
//
////////////////////////////////////////////////////////////////////////////////


int main( int argc, char **argv)
{
    // (0) welcome;
    char welcome[] =
        "\n--------------------------------------------------------------------------------\n"
        "MMEEP: Marquette MEEP, Version 1.0, Compiled "__DATE__" \n"
        "Design space exploration (DSE) via simulated annealing (SA) for patch antenna \n"
        "Cristinel Ababei & James E. Richie \n"
        "--------------------------------------------------------------------------------\n";
    printf("%s", welcome);

    IMPLEMENTATION_TYPE this_implementation_type;
    string this_in_file_name; // params filename;
    string this_out_file_name;
    if ( argc == 1) {
        //this_implementation_type = MULTITHREADED;
        //this_in_file_name = "Input.txt";
        //this_out_file_name = "Output.txt";
        printf("\nUsage: mmeep InFileName SEQUENTIAL/MULTITHREADED OutFileName (in that order!)\n\n");
        printf(" Option\t\tDescription. (Default hardcoded value)\n");
        printf(" InFileName\tRequired name of file with parameters. (Input.txt)\n");
        printf("           \tSee README.txt for description of file format.\n");
        printf(" SEQ/MUL\tType of implementation. Must be SEQUENTIAL, MULTITHREADED. (MULTITHREADED)\n");
        printf(" OutFileName\tName of file to save intermediate explored solutions. (Output.txt)\n");
        exit(1);
    } else if ( argc == 2) {
        this_in_file_name = argv[1];
        //printf("\nName of params file is: %s", this_in_file_name.c_str());
        this_implementation_type = MULTITHREADED;
        this_out_file_name = "Output.txt";
    } else if ( argc == 4) {
        this_in_file_name = argv[1];
        if (strcmp (argv[2], "SEQUENTIAL") == 0) {
            this_implementation_type = SEQUENTIAL;
        } else if (strcmp (argv[2], "MULTITHREADED") == 0) {
            this_implementation_type = MULTITHREADED;
        } else {
            this_implementation_type = MULTITHREADED; // default;
        }
        this_out_file_name = argv[3];
    } 


    // (1) runtime;
    // cpu time: method 1;
    clock_t start_clock, end_clock;
    clock_t diff_clock;
    start_clock = clock();
    assert(start_clock != (clock_t)(-1));
    // cpu time: method 2;
    //struct tms t1, t2;
    //times(&t1);
    // wall clock time;
    timeval start_wall, end_wall; // sec and microsec;
    gettimeofday( &start_wall, 0);



    // (2) create main application object and parse in users' variables 
    // that we'll sweep;
    APPLICATION_SIMULATED_ANNEALING_DSE application_dse;
    application_dse.set_implementation_type( this_implementation_type);
    application_dse.set_input_filename( this_in_file_name);
    application_dse.set_output_filename( this_out_file_name);
    application_dse.read_parameters_from_file();
    application_dse.print_parameters();
    

    // (3) create the mtps object, which implements the generator-workers 
    // multithreading framework;
    MTPC mtpc( &application_dse);
    mtpc.run();



    // (4) entertain user;
    printf("\n\nRESULTS SUMMARY:\n================\n");

    // best found solution and the values of costs of all explored solutions during SA;
    application_dse.print_best_found_solution();
    

    // runtime: cpu and wall times;
    printf("\n\nRUNTIME:\n================\n");
    // cpu time: method 1;
    end_clock = clock();
    assert(end_clock != (clock_t)(-1));
    diff_clock = end_clock - start_clock;
    printf("\nMMEEP total cputime  = %.3f sec\n",
        (double)diff_clock/CLOCKS_PER_SEC); // (processor time, clock())
    // cpu time: method 2;
    //times(&t2);
    //printf("vnoc total cputime = %.3f sec\n",
    //    (double)(t2.tms_utime-t1.tms_utime)/HZ); // (processor time, tms)
    // wall clock time;
    gettimeofday( &end_wall, 0);
    double diff_sec_usec = end_wall.tv_sec - start_wall.tv_sec + 
        double(end_wall.tv_usec - start_wall.tv_usec) / 1000000.0;
    printf ("MMEEP total walltime = %.3f sec\n\n", diff_sec_usec); // (wall clock time)


    // (5) clean-up;
    return 0;
}

